import { ok, throws } from 'assert';

import t from './globalContext';

describe('Type Constraints', function () {

  describe('TypeAlias TypeConstraints', function () {

    var EmailAddress = t.type('EmailAddress', t.string());

    it('should constrain a type', function () {
      EmailAddress.addConstraint(function (input) {
        if (input.length < 4) {
          return 'too short, cannot be a valid email address.';
        }
      }, function (input) {
        if (!/@/.test(input)) {
          return 'not a valid email address';
        }
      });
    });

    it('should accept valid input', function () {
      ok(EmailAddress.assert("foo@example.com"));
    });

    it('should reject invalid input: null', function () {
      throws(function () {
        return EmailAddress.assert(null);
      });
    });

    it('should reject invalid input: n@n', function () {
      throws(function () {
        return EmailAddress.assert('n@n');
      });
    });

    it('should reject invalid input: foo', function () {
      throws(function () {
        return EmailAddress.assert('foo');
      });
    });

    it('should reject invalid input: [1, 2, 3]', function () {
      throws(function () {
        return EmailAddress.assert([1, 2, 3]);
      });
    });
  });

  describe('ObjectTypeProperty Constraints', function () {

    var User = t.type('User', t.object({
      username: t.string()
    }));

    User.getProperty('username').addConstraint(function (input) {
      if (!/^([A-Z][A-Z0-9]*)$/i.test(input)) {
        return 'must be alphanumeric and start with a letter, no spaces.';
      }
    });

    it('should accept valid input', function () {
      User.assert({
        username: 'fine'
      });
    });

    it('should reject invalid input', function () {
      throws(function () {
        return User.assert({
          username: false
        });
      });
    });

    it('should reject more invalid input', function () {
      throws(function () {
        return User.assert({
          username: 'not fine'
        });
      });
    });
  });
});